from math import floor
from collections import Counter


def floored_percentage(val, digits):
    """
    Return string of floored value with given digits after period

    :param val:
    :param digits:
    :return:
    """
    val *= 10 ** (digits + 2)
    return '{1:.{0}f}%'.format(digits, floor(val) / 10 ** digits)


def pref_encoding(data, encoding_transform=None):
    """
    Return string with unique words in list with percentage of they frequency

    :param data:
    :param encoding_transform:
    :return str:
    """
    encoding_counted = Counter(data)
    total = sum(encoding_counted.values())
    sorted_encodings = sorted(encoding_counted.items(), key=lambda t: t[1], reverse=True)

    return ' / '.join(
            ["{:<1} [{:<4}]".format(encoding_transform(k) if encoding_transform else k,
                                    floored_percentage(v * 1.0 / total, 1)) for k, v in sorted_encodings])


def make_total_row(source, agg):
    """
    Execute agg column based function for source columns. For example if you need `total` in table data:

    Examples:
        src = [[1,1],[1,2],[1,3]]
        print(make_total_row(src, [sum, min]))
        >>> [3, 1]

    :param source:
    :param agg:
    :return:
    """
    return [agg[index](value) if callable(agg[index]) else agg[index] for index, value in enumerate(zip(*source))]



import statistics
import logging
from tqdm import tqdm
from rma.redis import *
from rma.helpers import pref_encoding, make_total_row, progress_iterator
from redis.exceptions import RedisError


class RealStringEntry(object):
    @staticmethod
    def get_int_encoded_bytes(redis, key_name):
        try:
            num_value = int(redis.get(key_name))
            if num_value < REDIS_SHARED_INTEGERS:
                return 0
            else:
                return size_of_pointer_fn()
        except ValueError:
            pass

        return size_of_pointer_fn()

    def __init__(self, redis, info, use_debug=True):
        """
        :param key_name:
        :param RmaRedis redis:
        :return:
        """
        key_name = info["name"]
        self.encoding = info["encoding"]
        self.logger = logging.getLogger(__name__)

        if self.encoding == REDIS_ENCODING_ID_INT:
            self.useful_bytes = self.get_int_encoded_bytes(redis, key_name)
            self.free_bytes = 0
            self.aligned = size_of_aligned_string_by_size(self.useful_bytes, encoding=self.encoding)
        elif self.encoding == REDIS_ENCODING_ID_EMBSTR or self.encoding == REDIS_ENCODING_ID_RAW:
            if use_debug:
                sdslen_response = redis.debug_sdslen(key_name)
                self.useful_bytes = sdslen_response['val_sds_len']
                self.free_bytes = sdslen_response['val_sds_avail']
            else:
                self.useful_bytes = size_of_aligned_string_by_size(redis.strlen(key_name), self.encoding)
                self.free_bytes = 0
            # INFO Rewrite this to support Redis >= 3.2 sds dynamic header
            sds_len = 8 + self.useful_bytes + self.free_bytes + 1
            self.aligned = size_of_aligned_string_by_size(sds_len, encoding=self.encoding)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False


class ValueString(object):
    def __init__(self, redis):
        self.redis = redis
        self.logger = logging.getLogger(__name__)

    def analyze(self, keys, total=0):
        key_stat = {
            'headers': ['Match', "Count", "Useful", "Free", "Real", "Ratio", "Encoding", "Min", "Max", "Avg"],
            'data': []
        }

        progress = tqdm(total=total,
                        mininterval=1,
                        desc="Processing String patterns",
                        leave=False)

        use_debug_command = True
        for pattern, data in keys.items():
            used_bytes = []
            free_bytes = []
            aligned_bytes = []
            encodings = []

            for key_info in progress_iterator(data, progress):
                try:
                    with RealStringEntry(redis=self.redis, info=key_info, use_debug=use_debug_command) as stat:
                        used_bytes.append(stat.useful_bytes)
                        free_bytes.append(stat.free_bytes)
                        aligned_bytes.append(stat.aligned)
                        encodings.append(stat.encoding)
                except RedisError as e:
                    # This code works in real time so key me be deleted and this code fail
                    error_string = repr(e)
                    self.logger.warning(error_string)
                    if 'DEBUG' in error_string:
                        use_debug_command = False

            total_elements = len(used_bytes)
            used_user = sum(used_bytes)
            free_user = sum(free_bytes)
            aligned = sum(aligned_bytes)
            preferred_encoding = pref_encoding(encodings, redis_encoding_id_to_str)

            min_bytes = min(used_bytes)
            mean = statistics.mean(used_bytes) if total_elements > 1 else min_bytes

            stat_entry = [
                pattern,
                total_elements,
                used_user,
                free_user,
                aligned,
                aligned / (used_user if used_user > 0 else 1),
                preferred_encoding,
                min_bytes,
                max(used_bytes),
                mean,
            ]
            key_stat['data'].append(stat_entry)

        key_stat['data'].sort(key=lambda e: e[1], reverse=True)
        key_stat['data'].append(make_total_row(key_stat['data'], ['Total:', sum, sum, 0, sum, 0, '', 0, 0, 0]))

        progress.close()

        return [
            "String value stat",
            key_stat
        ]



import sys
import fnmatch
import logging

from rma.redis import RmaRedis
from rma.scanner import Scanner
from rma.splitter import SimpleSplitter
from rma.redis_types import *
from rma.rule import *
from rma.reporters import *
from rma.helpers import floored_percentage

from collections import defaultdict
from redis.exceptions import ResponseError


def connect_to_redis(host, port, db=0, password=None):
    """

    :param host:
    :param port:
    :param db:
    :param password:
    :return RmaRedis:
    """
    try:
        redis = RmaRedis(host=host, port=port, db=db, password=password)
        if not check_redis_version(redis):
            sys.stderr.write('This script only works with Redis Server version 2.6.x or higher\n')
            sys.exit(-1)
    except ConnectionError as e:
        sys.stderr.write('Could not connect to Redis Server : %s\n' % e)
        sys.exit(-1)
    except ResponseError as e:
        sys.stderr.write('Could not connect to Redis Server : %s\n' % e)
        sys.exit(-1)
    return redis


def check_redis_version(redis):
    server_info = redis.info()
    version_str = server_info['redis_version']
    version = tuple(map(int, version_str.split('.')))

    if version[0] > 2 or (version[0] == 2 and version[1] >= 6):
        return True
    else:
        return False


class RmaApplication(object):
    globals = []

    types_rules = {
        REDIS_TYPE_ID_STRING: [],
        REDIS_TYPE_ID_HASH: [],
        REDIS_TYPE_ID_LIST: [],
        REDIS_TYPE_ID_SET: [],
        REDIS_TYPE_ID_ZSET: [],
    }

    def __init__(self, host="127.0.0.1", port=6367, password=None, db=0, match="*", limit=0, filters=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)

        self.splitter = SimpleSplitter()
        self.reporter = TextReporter()
        self.redis = connect_to_redis(host=host, port=port, db=db, password=password)

        self.match = match
        self.limit = limit if limit != 0 else sys.maxsize

        if 'types' in filters:
            self.types = list(map(redis_type_to_id, filters['types']))
        else:
            self.types = REDIS_TYPE_ID_ALL

        if 'behaviour' in filters:
            self.behaviour = filters['behaviour']
        else:
            self.behaviour = 'all'

    def init_globals(self, redis):
        self.globals.append(GlobalKeySpace(redis=redis))

    def init_types_rules(self, redis):
        self.types_rules[REDIS_TYPE_ID_STRING].append(KeyString(redis=redis))
        self.types_rules[REDIS_TYPE_ID_STRING].append(ValueString(redis=redis))
        self.types_rules[REDIS_TYPE_ID_HASH].append(KeyString(redis=redis))
        self.types_rules[REDIS_TYPE_ID_HASH].append(Hash(redis=redis))
        self.types_rules[REDIS_TYPE_ID_LIST].append(KeyString(redis=redis))
        self.types_rules[REDIS_TYPE_ID_LIST].append(List(redis=redis))

        self.types_rules[REDIS_TYPE_ID_SET].append(KeyString(redis=redis))
        self.types_rules[REDIS_TYPE_ID_SET].append(Set(redis=redis))

        self.types_rules[REDIS_TYPE_ID_ZSET].append(KeyString(redis=redis))

    def run(self):
        self.init_types_rules(redis=self.redis)
        self.init_globals(redis=self.redis)

        str_res = []
        is_all = self.behaviour == 'all'
        with Scanner(redis=self.redis, match=self.match, accepted_types=self.types) as scanner:
            keys = defaultdict(list)
            for v in scanner.scan(limit=self.limit):
                keys[v["type"]].append(v)

            print("\r\nAggregating keys by pattern and type")
            keys = {k: self.get_pattern_aggregated_data(v) for k, v in keys.items()}

            print("\r\nApply rules")
            if self.behaviour == 'global' or is_all:
                str_res += self.do_globals()
            if self.behaviour == 'scanner' or is_all:
                str_res += self.do_scanner(self.redis, keys)
            if self.behaviour == 'ram' or is_all:
                str_res += self.do_ram(keys)

        self.reporter.print(str_res)

    def do_globals(self):
        res = ['Server stat']
        for glob in self.globals:
            res += glob.analyze()

        return res

    def do_scanner(self, r, res):
        key_stat = {
            'headers': ['Match', "Count", "Type", "%"],
            'data': []
        }
        total = min(r.dbsize(), self.limit)
        for key, aggregate_patterns in res.items():
            self.logger.debug("Processing type %s" % type_id_to_redis_type(key))
            r_type = type_id_to_redis_type(key)

            for k, v in aggregate_patterns.items():
                key_stat['data'].append([k, len(v), r_type, floored_percentage(len(v) / total, 2)])

        key_stat['data'].sort(key=lambda x: x[1], reverse=True)
        return ["Keys by types", key_stat]

    def do_ram(self, res):
        ret = []
        for key, aggregate_patterns in res.items():
            self.logger.debug("Processing type %s" % type_id_to_redis_type(key))
            if key in self.types_rules and key in self.types:
                ret.append("Processing {0}".format(type_id_to_redis_type(key)))
                for rule in self.types_rules[key]:
                    total_keys = sum(len(values) for key, values in aggregate_patterns.items())
                    ret += (rule.analyze(keys=aggregate_patterns, total=total_keys))

        return ret

    def get_pattern_aggregated_data(self, data):
        split_patterns = self.splitter.split((obj["name"] for obj in data))
        self.logger.debug(split_patterns)

        aggregate_patterns = {item: [] for item in split_patterns}
        for pattern in split_patterns:
            aggregate_patterns[pattern] = list(filter(lambda obj: fnmatch.fnmatch(obj["name"], pattern), data))

        return aggregate_patterns



num_version = (0, 1, 11)
version = '.'.join([str(num) for num in num_version])

classifiers = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python'
]



#!/usr/bin/env python
import time
import logging
from argparse import ArgumentParser, HelpFormatter
from rma.application import RmaApplication

logging.basicConfig(level=logging.INFO)

VALID_TYPES = ("string", "hash", "list", "set", "zset")
VALID_MODES = ('all', 'scanner', 'ram', 'global')


class CustomHelpFormatter(HelpFormatter):
    def __init__(self, prog):
        super().__init__(prog, max_help_position=40, width=80)

    def _format_action_invocation(self, action):
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)
        return ', '.join(action.option_strings) + ' ' + args_string


def parser_formatter(prog):
    return CustomHelpFormatter(prog)


def main():
    description = """RMA is used to scan Redis key space in and aggregate memory usage statistic by key patterns."""

    parser = ArgumentParser(prog='rma', description=description, formatter_class=parser_formatter)
    parser.add_argument("-s", "--server",
                        dest="host",
                        default="127.0.0.1",
                        help="Redis Server hostname. Defaults to 127.0.0.1")
    parser.add_argument("-p", "--port",
                        dest="port",
                        default=6379,
                        type=int,
                        help="Redis Server port. Defaults to 6379")
    parser.add_argument("-a", "--password",
                        dest="password",
                        help="Password to use when connecting to the server")
    parser.add_argument("-d", "--db",
                        dest="db",
                        default=0,
                        help="Database number, defaults to 0")
    parser.add_argument("-m", "--match",
                        dest="match",
                        default="*",
                        help="Keys pattern to match")
    parser.add_argument("-l", "--limit",
                        dest="limit",
                        default="0",
                        type=int,
                        help="Get max key matched by pattern")
    parser.add_argument("-b", "--behaviour",
                        dest="behaviour",
                        default="all",
                        help="Specify application working mode. Allowed values are " + ', '.join(VALID_MODES))
    parser.add_argument("-t", "--type",
                        dest="types",
                        action="append",
                        help="""Data types to include. Possible values are string, hash, list, set.
                              Multiple types can be provided. If not specified, all data types will be returned.
                              Allowed values are """ + ', '.join(VALID_TYPES))

    options = parser.parse_args()

    filters = {}
    if options.behaviour:
        if options.behaviour not in VALID_MODES:
            raise Exception(
                    'Invalid behaviour provided - %s. Expected one of %s' % (
                        options.behaviour, (", ".join(VALID_TYPES))))
        else:
            filters['behaviour'] = options.behaviour

    if options.types:
        filters['types'] = []
        for x in options.types:
            if x not in VALID_TYPES:
                raise Exception('Invalid type provided - %s. Expected one of %s' % (x, (", ".join(VALID_TYPES))))
            else:
                filters['types'].append(x)

    app = RmaApplication(host=options.host, port=options.port, db=options.db,
                         password=options.password, match=options.match, limit=options.limit, filters=filters)

    start_time = time.clock()
    app.run()
    print("\r\nDone in %s seconds" % (time.clock() - start_time))


if __name__ == '__main__':
    main()



def is_num(value):
    try:
        int(value)
        return True
    except ValueError:
        return False


# http://code.activestate.com/recipes/415233-getting-minmax-in-a-sequence-greaterless-than-some/
def min_ge(seq, val):
    """
    Same as min_gt() except items equal to val are accepted as well.

    :param seq:
    :param val:
    :return:

    Examples:
        min_ge([1, 3, 6, 7], 6)
        >>>6
        min_ge([2, 3, 4, 8], 8)
        >>>8
    """

    for v in seq:
        if v >= val:
            return v
    return None


def next_power_of_2(n):
    """
    Return next power of 2 greater than or equal to n
    """
    return 2**(n - 1).bit_length()


def is_power2(num):
    """
    states if a number is a power of two
    """
    return num != 0 and ((num & (num - 1)) == 0)



from redis import StrictRedis
from rma.helpers import is_num, next_power_of_2
from rma.jemalloc import Jemalloc
from rma.redis_types import *


size_of_pointer = 8
new_sds = False
allocator = Jemalloc

# Const
REDIS_SHARED_INTEGERS = 10000
REDIS_ENCODING_EMBSTR_SIZE_LIMIT = 39


# String encodings
REDIS_ENCODING_INT = 'int'
REDIS_ENCODING_EMBSTR = 'embstr'
REDIS_ENCODING_RAW = 'raw'


def size_of_pointer_fn():
    """
    Return pointer size for current Redis connection
    :return int:
    """
    # UNDONE
    return 8


def get_redis_object_overhead():
    """
    See struct sdshdr over here https://github.com/antirez/redis/blob/unstable/src/sds.h:
    typedef struct redisObject {
        unsigned type:4;
        unsigned encoding:4;
        unsigned lru:REDIS_LRU_BITS; /* lru time (relative to server.lruclock) */
        int refcount;
        void *ptr;
    } robj;

    :return:
    """
    return 4 + 4 + 8 + 8 + size_of_pointer_fn()


def get_string_encoding(value):
    if is_num(value):
        return REDIS_ENCODING_ID_INT
    elif len(value) < REDIS_ENCODING_EMBSTR_SIZE_LIMIT:
        return REDIS_ENCODING_ID_EMBSTR
    else:
        return REDIS_ENCODING_ID_RAW


def size_of_sds_string(value, encoding=REDIS_ENCODING_ID_INT):
    if encoding == REDIS_ENCODING_ID_INT:
        try:
            num_value = int(value)
            if num_value < REDIS_SHARED_INTEGERS:
                return 0
            else:
                return size_of_pointer_fn()
        except ValueError:
            pass

    return 8 + len(value) + 1


def size_of_aligned_string(value, encoding=""):
    if encoding == "":
        encoding = get_string_encoding(value)

    str_len = size_of_sds_string(value, encoding)
    return size_of_aligned_string_by_size(str_len, encoding)


def size_of_aligned_string_by_size(sdslen, encoding):
    r_obj = get_redis_object_overhead()
    if encoding == REDIS_ENCODING_ID_INT:
        return allocator.align(r_obj)
    elif encoding == REDIS_ENCODING_ID_EMBSTR:
        return allocator.align(sdslen + r_obj)
    else:
        # REDIS_ENCODING_ID_RAW
        return allocator.align(sdslen) + allocator.align(r_obj)


def dict_overhead(size):
    return Jemalloc.align(56 + 2*size_of_pointer_fn() + next_power_of_2(size) * dict_entry_overhead())


def dict_entry_overhead():
    return 3*size_of_pointer_fn()


def ziplist_overhead(size):
    return Jemalloc.align(12 + 21 * size)


def size_of_ziplist_aligned_string(value):
    # Looks like we need something more complex here. We use calculation as 21 bytes per entry + len of string
    # or len of pointer. Redis use more RAM saving policy but after aligning it has infelicity ~3-5%
    try:
        num_value = int(value)
        return Jemalloc.align(size_of_pointer_fn())
    except ValueError:
        pass

    return Jemalloc.align(len(value))


def linkedlist_overhead():
    # A list has 5 pointers + an unsigned long
    return 8 + 5*size_of_pointer_fn()


def linkedlist_entry_overhead():
    # See https://github.com/antirez/redis/blob/unstable/src/adlist.h
    # A node has 3 pointers
    return 3*size_of_pointer_fn()


def size_of_linkedlist_aligned_string(value):
    return Jemalloc.align(linkedlist_entry_overhead() + len(value))


def intset_overhead(size):
    #     typedef struct intset {
    #     uint32_t encoding;
    #     uint32_t length;
    #     int8_t contents[];
    # } intset;
    return (4 + 4) * size


def intset_aligned(value):
    overhead = intset_overhead(1)
    try:
        val = int(value)

    except ValueError:
        return Jemalloc.align(8 + overhead)

    if val < 65535:
        return Jemalloc.align(2 + overhead)

    if val < 2147483647:
        return Jemalloc.align(4 + overhead)

    return Jemalloc.align(8 + overhead)


def parse_debug(response):
    """
    Parse the result of Redis's DEBUG command into a Python dict
    :param bytearray response:
    :return disc:
    """
    info = {}
    response = response.decode('utf8')

    for line in response.split(','):
        if line.find(':') != -1:
            key, value = line.split(':', 1)
            info[key.strip()] = int(value)
        else:
            # if the line isn't splittable, append it to the "__raw__" key
            info.setdefault('__raw__', []).append(line)

    return info


class RmaRedis(StrictRedis):
    def debug_sdslen(self, key):
        return parse_debug(self.execute_command("DEBUG SDSLEN", key))



from itertools import tee
from tqdm import tqdm
from rma.redis import *
from rma.helpers import pref_encoding, make_total_row, progress_iterator

import statistics


class SetStatEntry(object):
    def __init__(self, info, redis):
        """
        :param key_name:
        :param RmaRedis redis:
        :return:
        """

        key_name = info["name"]
        self.values = [v for v in redis.sscan_iter(key_name, '*', 1000)]
        self.encoding = info["encoding"]
        self.count = len(self.values)

        if self.encoding == REDIS_ENCODING_ID_HASHTABLE:
            self.system = dict_overhead(self.count)
            self.valueUsedBytes = sum((len(x) for x in self.values))
            self.valueAlignedBytes = sum(map(size_of_aligned_string, self.values))
            self.total = self.valueAlignedBytes + self.system
        elif self.encoding == REDIS_ENCODING_ID_INTSET:
            self.system = intset_overhead(self.count)
            # In case if intset valueAlignedBytes already contains system overhead
            self.valueAlignedBytes = sum(map(size_of_ziplist_aligned_string, self.values))
            self.valueUsedBytes = self.valueAlignedBytes
            self.total = self.valueAlignedBytes
        else:
            raise Exception('Panic', 'Unknown encoding %s in %s' % (self.encoding, key_name))


class SetAggregator(object):
    def __init__(self, all_obj, total):
        self.total_elements = total

        g00, g0, g3, v1, v2, v3 = tee(all_obj, 6)

        self.encoding = pref_encoding([obj.encoding for obj in g00], redis_encoding_id_to_str)
        self.system = sum(obj.system for obj in g0)
        if total > 1:
            self.fieldAvgCount = statistics.mean(obj.count for obj in g3)
        else:
            self.fieldAvgCount = min((obj.count for obj in g3))

        self.valueUsedBytes = sum(obj.valueUsedBytes for obj in v1)
        self.valueAlignedBytes = sum(obj.valueAlignedBytes for obj in v2)
        self.total = sum(obj.total for obj in v3)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False


class Set(object):
    def __init__(self, redis):
        """
        :param RmaRedis redis:
        :return:
        """
        self.redis = redis

    def analyze(self, keys, total=0):
        key_stat = {
            'headers': ['Match', "Count", "Avg Count", "Value mem", "Real", "Ratio", "System*", "Encoding", "Total"],
            'data': []
        }

        progress = tqdm(total=total,
                        mininterval=1,
                        desc="Processing Set patterns",
                        leave=False)

        for pattern, data in keys.items():
            agg = SetAggregator(progress_iterator((SetStatEntry(x, self.redis) for x in data), progress), len(data))

            stat_entry = [
                pattern,
                len(data),
                agg.fieldAvgCount,
                agg.valueUsedBytes,
                agg.valueAlignedBytes,
                agg.valueAlignedBytes / (agg.valueUsedBytes if agg.valueUsedBytes > 0 else 1),
                agg.system,
                agg.encoding,
                agg.total
            ]

            key_stat['data'].append(stat_entry)

        key_stat['data'].sort(key=lambda x: x[8], reverse=True)
        key_stat['data'].append(make_total_row(key_stat['data'], ['Total:', sum, 0, sum, sum, 0, sum, '', sum]))

        progress.close()

        return [
            "SET stat",
            key_stat
        ]



import logging
from rma.redis import dict_overhead
from redis.exceptions import ResponseError


class GlobalKeySpace(object):
    def __init__(self, redis):
        """
        :param RmaRedis redis:
        :return:
        """
        self.redis = redis
        self.logger = logging.getLogger(__name__)

    def analyze(self):
        total_keys = self.redis.dbsize()

        keys_ = [
            ["Total keys in db", total_keys],
            ["RedisDB key space overhead", dict_overhead(total_keys)]
        ]
        try:
            keys_ += [["Used `{0}`".format(key), value] for key, value in self.redis.config_get("*max-*-*").items()]
        except ResponseError as e:
            self.logger.warning("*max* option skipped: %s", repr(e))

        keys_ += [["Info `{0}`".format(key), value] for key, value in self.redis.info('memory').items()]

        return [
            {
                'headers': ['Stat', "Value"],
                'data': keys_
            }
        ]



import statistics
from itertools import tee
from tqdm import tqdm
from rma.redis import *
from rma.helpers import pref_encoding, make_total_row, progress_iterator


class ListStatEntry(object):
    def __init__(self, info, redis):
        """
        :param key_name:
        :param RmaRedis redis:
        :return:
        """
        key_name = info["name"]
        self.encoding = info['encoding']

        self.values = redis.lrange(key_name, 0, -1)
        self.count = len(self.values)
        import time
        time.sleep(0.001)
        used_bytes_iter, min_iter, max_iter = tee((len(x) for x in self.values), 3)

        if self.encoding == REDIS_ENCODING_ID_LINKEDLIST:
            self.system = dict_overhead(self.count)
            self.valueAlignedBytes = sum(map(size_of_linkedlist_aligned_string, self.values))
        elif self.encoding == REDIS_ENCODING_ID_ZIPLIST or self.encoding == REDIS_ENCODING_ID_QUICKLIST:
            # Undone `quicklist`
            self.system = ziplist_overhead(self.count)
            self.valueAlignedBytes = sum(map(size_of_ziplist_aligned_string, self.values))
        else:
            raise Exception('Panic', 'Unknown encoding %s in %s' % (self.encoding, key_name))

        self.valueUsedBytes = sum(used_bytes_iter)

        if self.count > 0:
            self.valueMin = min(min_iter)
            self.valueMax = max(max_iter)
        else:
            self.valueMin = 0
            self.valueMax = 0


class ListAggregator(object):
    def __init__(self, all_obj, total):
        self.total_elements = total

        encode_iter, sys_iter, avg_iter, stdev_iter, min_iter, max_iter, value_used_iter, value_align_iter = \
            tee(all_obj, 8)

        self.encoding = pref_encoding([obj.encoding for obj in encode_iter], redis_encoding_id_to_str)
        self.system = sum(obj.system for obj in sys_iter)
        if total > 1:
            self.fieldAvgCount = statistics.mean(obj.count for obj in avg_iter)
            self.fieldStdev = statistics.stdev(obj.count for obj in stdev_iter)
            self.fieldMinCount = min((obj.count for obj in min_iter))
            self.fieldMaxCount = max((obj.count for obj in max_iter))
        else:
            self.fieldAvgCount = min((obj.count for obj in avg_iter))
            self.fieldStdev = 0
            self.fieldMinCount = self.fieldAvgCount
            self.fieldMaxCount = self.fieldAvgCount

        self.valueUsedBytes = sum(obj.valueUsedBytes for obj in value_used_iter)
        self.valueAlignedBytes = sum(obj.valueAlignedBytes for obj in value_align_iter)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False


class List(object):
    def __init__(self, redis):
        """
        :param RmaRedis redis:
        :return:
        """
        self.redis = redis

    def analyze(self, keys, total=0):
        key_stat = {
            'headers': ['Match', "Count", "Avg Count", "Min Count", "Max Count", "Stdev Count", "Value mem", "Real", "Ratio", "System", "Encoding", "Total"],
            'data': []
        }

        progress = tqdm(total=total,
                        mininterval=1,
                        desc="Processing List patterns",
                        leave=False)

        for pattern, data in keys.items():
            agg = ListAggregator(progress_iterator((ListStatEntry(x, self.redis) for x in data), progress), len(data))

            stat_entry = [
                pattern,
                len(data),
                agg.fieldAvgCount,
                agg.fieldMinCount,
                agg.fieldMaxCount,
                agg.fieldStdev,
                agg.valueUsedBytes,
                agg.valueAlignedBytes,
                agg.valueAlignedBytes / (agg.valueUsedBytes if agg.valueUsedBytes > 0 else 1),
                agg.system,
                agg.encoding,
                agg.valueAlignedBytes + agg.system
            ]

            key_stat['data'].append(stat_entry)
            progress.update()

        key_stat['data'].sort(key=lambda x: x[8], reverse=True)
        key_stat['data'].append(make_total_row(key_stat['data'], ['Total:', sum, 0, 0, 0, 0, sum, sum, 0, sum, '', sum]))

        progress.close()
        return [
            "List stat",
            key_stat
        ]



from rma.helpers import is_power2, min_ge


class Jemalloc(object):
    """
    Small: All 2^n-aligned allocations of size 2^n will incur no additional overhead, due to how small allocations are
    aligned and packed.

    Small: [8], [16, 32, 48, ..., 128], [192, 256, 320, ..., 512], [768, 1024, 1280, ..., 3840]

    Large: The worst case size is half the chunk size, in which case only one allocation per chunk can be allocated.
    If the remaining (nearly) half of the chunk isn't otherwise useful for smaller allocations, the overhead will
    essentially be 50%. However, assuming you use a diverse mixture of size classes, the actual overhead shouldn't be a
    significant issue in practice.

    Large: [4 KiB, 8 KiB, 12 KiB, ..., 4072 KiB]

    Huge: Extra virtual memory is mapped, then the excess is trimmed and unmapped.  This can leave virtual memory holes,
    but it incurs no physical memory overhead.  Earlier versions of jemalloc heuristically attempted to optimistically
    map chunks without excess that would need to be trimmed, but it didn't save much system call overhead in practice.

    Huge: [4 MiB, 8 MiB, 12 MiB, ..., 512 MiB]
    """
    @staticmethod
    def align(size):
        if size <= 4096:
            # Small
            if is_power2(size):
                return size
            elif size < 128:
                return min_ge(range(16, 128 + 1, 16), size)
            elif size < 512:
                return min_ge(range(192, 512 + 1, 64), size)
            else:
                return min_ge(range(768, 4096 + 1, 256), size)
        elif size < 4194304:
            # Large
            return min_ge(range(4096, 4194304 + 1, 4096), size)
        else:
            # Huge
            return min_ge(range(4194304, 536870912 + 1, 4194304), size)



from tabulate import tabulate


class TextReporter:
    def print(self, data):
        for out in data:
            print("\r\n")
            if isinstance(out, str):
                print(out)
            else:
                print(tabulate(out["data"], headers=out["headers"], floatfmt=".2f", tablefmt="pipe"))



def has_numbers(value):
    return any(char.isdigit() for char in value)


def dict_build(indict, pre=None):
    pre = pre[:] if pre else []
    if len(indict):
        for key, value in indict.items():
            if len(value):
                for d in dict_build(value, pre=pre+[key]):
                    yield d
            else:
                yield pre + [key]
    else:
        yield pre


def map_part_to_glob(index, part):
    if index == 0:
        return part

    if not part or has_numbers(part):
        return '*'

    return part


class SimpleSplitter(object):
    def split(self, data, separator=":"):
        pass1 = map(lambda x: list(map_part_to_glob(i, y) for i, y in enumerate(x.split(separator))), data)
        pass2 = self.fold_to_tree(pass1)
        return self.unfold_to_list(pass2, separator)

    def fold_to_tree(self, pass1):
        tree = {}
        for item in pass1:
            t_len = len(item)
            if t_len not in tree:
                tree[t_len] = {}

            subtree = tree[t_len]
            deep = 0
            for part in item:
                deep += 1
                if '*' in subtree:
                    subtree = subtree['*']
                    continue

                if part in subtree:
                    subtree = subtree[part]
                    continue

                subtree[part] = {}
                if part != '*':
                    subtree = subtree[part]
                    continue

                if len(subtree) == 1:
                    continue

                if deep > 1:
                    self.merge_subtree(subtree)

                subtree = subtree[part]
        return tree

    @staticmethod
    def merge_subtree(subtree):
        all_keys = []
        for sub_part in subtree.keys():
            if sub_part != '*':
                all_keys.append(sub_part)
                subtree['*'].update(subtree[sub_part])
        for sub_part in all_keys:
            del subtree[sub_part]

    @staticmethod
    def unfold_to_list(tree, separator):
        res = set()

        for sub_tree in tree.values():
            for compound_key in dict_build(sub_tree):
                res.add(separator.join(compound_key))
        return res



import statistics
from tqdm import tqdm
from itertools import tee
from rma.redis import *
from rma.helpers import pref_encoding, make_total_row, progress_iterator


class StringEntry(object):
    def __init__(self, value=""):
        self.encoding = get_string_encoding(value)
        self.useful_bytes = len(value)
        self.free_bytes = 0
        self.aligned = size_of_aligned_string(value, encoding=self.encoding)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False


class KeyString(object):
    def __init__(self, redis):
        """
        :param RmaRedis redis:
        :return:
        """
        self.redis = redis

    def analyze(self, keys, total=0):
        """

        :param keys:
        :param progress:
        :return:
        """
        key_stat = {
            'headers': ['Match', "Count", "Useful", "Real", "Ratio", "Encoding", "Min", "Max", "Avg"],
            'data': []
        }

        progress = tqdm(total=total,
                        mininterval=1,
                        desc="Processing keys",
                        leave=False)

        for pattern, data in keys.items():
            used_bytes_iter, aligned_iter, encoding_iter = tee(
                    progress_iterator((StringEntry(value=x["name"]) for x in data), progress), 3)

            total_elements = len(data)
            aligned = sum(obj.aligned for obj in aligned_iter)
            used_bytes_generator = (obj.useful_bytes for obj in used_bytes_iter)
            useful_iter, min_iter, max_iter, mean_iter = tee(used_bytes_generator, 4)

            prefered_encoding = pref_encoding((obj.encoding for obj in encoding_iter), redis_encoding_id_to_str)
            min_value = min(min_iter)
            if total_elements < 2:
                avg = min_value
            else:
                avg = statistics.mean(mean_iter)

            used_user = sum(useful_iter)

            stat_entry = [
                pattern, total_elements, used_user, aligned, aligned / used_user, prefered_encoding,
                min_value, max(max_iter), avg,
            ]
            key_stat['data'].append(stat_entry)

        key_stat['data'].sort(key=lambda x: x[1], reverse=True)
        key_stat['data'].append(make_total_row(key_stat['data'], ['Total:', sum, sum, sum, 0, '', 0, 0, 0]))

        progress.close()

        return [
            "key stats",
            key_stat
        ]



import logging
import msgpack
from redis.exceptions import ResponseError
from tqdm import tqdm
from rma.redis import *


def chunker(seq, size):
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))


class Scanner(object):
    """
    Get all keys from Redis database with given match and limits. If limit specified would be retrieved not more then
    limit keys.
    """
    logger = logging.getLogger(__name__)

    def __init__(self, redis, match="*", accepted_types=None):
        """
        :param RmaRedis redis:
        :param str match: Wild card match supported in Redis SCAN command
        :return:
        """
        self.redis = redis
        self.match = match
        self.accepted_types = accepted_types[:] if accepted_types else REDIS_TYPE_ID_ALL
        self.pipeline_mode = False
        self.resolve_types_script = self.redis.register_script("""
            local ret = {}
            for i = 1, #KEYS do
                local type = redis.call("TYPE", KEYS[i])
                local encoding = redis.call("OBJECT", "ENCODING",KEYS[i])
                ret[i] = {type["ok"], encoding}
            end
            return cmsgpack.pack(ret)
        """)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def batch_scan(self, count=1000, batch_size=3000):
        ret = []
        for key in self.redis.scan_iter(self.match, count=count):
            ret.append(key)
            if len(ret) == batch_size:
                yield from self.resolve_types(ret)

        if len(ret):
            yield from self.resolve_types(ret)

    def resolve_types(self, ret):
        if not self.pipeline_mode:
            try:
                key_with_types = msgpack.unpackb(self.resolve_types_script(ret))
            except ResponseError as e:
                if "CROSSSLOT" not in repr(e):
                    raise e
                key_with_types = self.resolve_with_pipe(ret)
                self.pipeline_mode = True
        else:
            key_with_types = self.resolve_with_pipe(ret)

        for i in range(0, len(ret)):
            yield key_with_types[i], ret[i]

        ret.clear()

    def resolve_with_pipe(self, ret):
        pipe = self.redis.pipeline(transaction=False)
        for key in ret:
            pipe.type(key)
            pipe.object('ENCODING', key)
        key_with_types = [{'type': x, 'encoding': y} for x, y in chunker(pipe.execute(), 2)]
        return key_with_types

    def scan(self, limit=1000):
        with tqdm(total=min(limit, self.redis.dbsize()), desc="Match {0}".format(self.match),
                  miniters=1000) as progress:

            total = 0
            for key_tuple in self.batch_scan():
                key_info, key_name = key_tuple
                key_type, key_encoding = key_info
                if not key_name:
                    self.logger.warning(
                        '\r\nWarning! Scan iterator return key with empty name `` and type %s', key_type)
                    continue

                to_id = redis_type_to_id(key_type)
                if to_id in self.accepted_types:
                    key_info_obj = {
                        'name': key_name.decode("utf-8"),
                        'type': to_id,
                        'encoding': redis_encoding_str_to_id(key_encoding)
                    }
                    yield key_info_obj

                progress.update()

                total += 1
                if total > limit:
                    logging.info("\r\nLimit %s reached", limit)
                    break



#!/usr/bin/env python
import os
from os.path import join

base_dir = os.path.dirname(__file__)
readme_path = join(base_dir, 'README.rst')
changes = join(base_dir, 'CHANGES')

long_description = open(readme_path).read() + '\n' + open(changes).read()

__pkginfo__ = {}
with open(os.path.join(base_dir, "rma", "__pkginfo__.py")) as f:
    exec(f.read(), __pkginfo__)

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(name='rma',
      version=__pkginfo__['version'],
      description='Utilities to profile Redis RAM usage',
      long_description=long_description,
      url='https://github.com/gamenet/redis-memory-analyzer',
      author='Nikolay Bondarenko',
      author_email='misterionkell@gmail.com',
      maintainer='Nikolay Bondarenko',
      maintainer_email='misterionkell@gmail.com',
      keywords=['Redis', 'Memory Profiler'],
      license='MIT',
      install_requires=['redis', 'tabulate', 'tqdm', 'msgpack-python'],
      include_package_data=True,
      packages=['rma', 'rma.helpers', 'rma.reporters', 'rma.rule', 'rma.cli'],
      package_data={
          'rma.cli': ['*.template']
      },
      test_suite='tests.all_tests',
      entry_points={
          'console_scripts': [
              'rma = rma.cli.rma_cli:main',
          ],
      },
      classifiers=__pkginfo__['classifiers'],)



from rma.helpers.formating import floored_percentage, pref_encoding, make_total_row
from rma.helpers.numbers import is_num, min_ge, next_power_of_2, is_power2


def progress_iterator(x, progress):
    for i in x:
        yield i
        progress.update()


__all__ = ['floored_percentage', 'pref_encoding', 'make_total_row', 'is_num', 'min_ge', 'next_power_of_2', 'is_power2',
           'progress_iterator']



# Internal encoding mapping
REDIS_ENCODING_ID_RAW = 0
REDIS_ENCODING_ID_INT = 1
REDIS_ENCODING_ID_EMBSTR = 2
REDIS_ENCODING_ID_HASHTABLE = 3
REDIS_ENCODING_ID_ZIPLIST = 4
REDIS_ENCODING_ID_LINKEDLIST = 5
REDIS_ENCODING_ID_QUICKLIST =6
REDIS_ENCODING_ID_INTSET = 7
REDIS_ENCODING_ID_SKIPLIST = 8
REDIS_ENCODING_ID_ALL = [
    REDIS_ENCODING_ID_RAW,
    REDIS_ENCODING_ID_INT,
    REDIS_ENCODING_ID_EMBSTR,
    REDIS_ENCODING_ID_HASHTABLE,
    REDIS_ENCODING_ID_ZIPLIST,
    REDIS_ENCODING_ID_LINKEDLIST,
    REDIS_ENCODING_ID_QUICKLIST,
    REDIS_ENCODING_ID_INTSET,
    REDIS_ENCODING_ID_SKIPLIST
]

# Internal type mapping
REDIS_TYPE_ID_UNKNOWN = -1
REDIS_TYPE_ID_STRING = 0
REDIS_TYPE_ID_HASH = 1
REDIS_TYPE_ID_LIST = 2
REDIS_TYPE_ID_SET = 3
REDIS_TYPE_ID_ZSET = 4
REDIS_TYPE_ID_ALL = [
    REDIS_TYPE_ID_STRING,
    REDIS_TYPE_ID_HASH,
    REDIS_TYPE_ID_LIST,
    REDIS_TYPE_ID_SET,
    REDIS_TYPE_ID_ZSET
]

REDIS_ENCODING_STR_TO_ID_LIB = {
    b'raw': REDIS_ENCODING_ID_RAW,
    b'int': REDIS_ENCODING_ID_INT,
    b'embstr': REDIS_ENCODING_ID_EMBSTR,
    b'hashtable': REDIS_ENCODING_ID_HASHTABLE,
    b'ziplist': REDIS_ENCODING_ID_ZIPLIST,
    b'linkedlist': REDIS_ENCODING_ID_LINKEDLIST,
    b'quicklist': REDIS_ENCODING_ID_QUICKLIST,
    b'intset': REDIS_ENCODING_ID_INTSET,
    b'skiplist': REDIS_ENCODING_ID_SKIPLIST,
}

REDIS_ENCODING_ID_TO_STR_LIB = dict((v, k) for k, v in REDIS_ENCODING_STR_TO_ID_LIB.items())


def redis_encoding_str_to_id(key_encoding):
    if key_encoding in REDIS_ENCODING_STR_TO_ID_LIB:
        return REDIS_ENCODING_STR_TO_ID_LIB[key_encoding]

    raise ValueError("Invalid encoding `%s` given" % key_encoding)


def redis_encoding_id_to_str(key_encoding):
    if key_encoding in REDIS_ENCODING_ID_TO_STR_LIB:
        return REDIS_ENCODING_ID_TO_STR_LIB[key_encoding].decode('utf8')

    raise ValueError("Invalid encoding `%s` given" % key_encoding)


def redis_type_to_id(key_type):
    """
    Convert redis type string to internal num type
    :param str key_type:
    :return int:
    """
    if key_type == b'string' or key_type == 'string':
        return REDIS_TYPE_ID_STRING
    elif key_type == b'hash' or key_type == 'hash':
        return REDIS_TYPE_ID_HASH
    elif key_type == b'list' or key_type == 'list':
        return REDIS_TYPE_ID_LIST
    elif key_type == b'set' or key_type == 'set':
        return REDIS_TYPE_ID_SET
    elif key_type == b'zset' or key_type == 'zset':
        return REDIS_TYPE_ID_ZSET
    else:
        return REDIS_TYPE_ID_UNKNOWN


def type_id_to_redis_type(type_id):
    """
    Convert internal type id to Redis string type
    :param int type_id:
    :return str:
    """
    if type_id == REDIS_TYPE_ID_STRING:
        return 'string'
    elif type_id == REDIS_TYPE_ID_HASH:
        return 'hash'
    elif type_id == REDIS_TYPE_ID_LIST:
        return 'list'
    elif type_id == REDIS_TYPE_ID_SET:
        return 'set'
    elif type_id == REDIS_TYPE_ID_ZSET:
        return 'zset'
    else:
        return 'unknown'


from rma.rule.KeyString import KeyString
from rma.rule.ValueString import ValueString
from rma.rule.GlobalKeySpace import GlobalKeySpace
from rma.rule.Hash import Hash
from rma.rule.List import List
from rma.rule.Set import Set

__all__ = ['KeyString', 'ValueString', 'GlobalKeySpace', 'Hash', 'List', 'Set']
